﻿var _ = require('lodash');
var cp = require('child_process');
var os = require('os');
var fs = require('fs');

var MAC = 'HWaddr';
var INET = 'inet';
var BCAST = 'Bcast';
var RUNNING = 'RUNNING'; var UNPLUGGED = 'UNPLUGGED';
var UP = 'UP'; var DOWN = 'DOWN';

// Globals
//#region Globals
var NetworkManagerIo;
var NetworkManagerLogger;
var NetworkManagerNetSnmp;

//#endregion Globals

//----------------------------------------------------------
// Network Interfaces
//#region Network Interfaces

exports.InitializeNetworkInterfaces = function (db, logger, netSnmp) {
    NetworkManagerDb = db;
    NetworkManagerLogger = logger;
    NetworkManagerNetSnmp = netSnmp;
}

//#region Get Network Interfaces

// Return a list of network interfaces 
exports.GetNetworkInterfaces = function (callback) {
    
    // Step 1: Get basic IfConfig information
    try {
        cp.exec('/sbin/ifconfig', function (err, ifConfigOut, stderr) {
            if (err || stderr) {
                NetworkManagerLogger.error('NetworkManager.GetNetworkInterfaces: ifconfig err=' + err + ' stderr=' + stderr);
            } else {
                NetworkManagerLogger.info('NetworkManager.GetNetworkInterfaces: raw ifconfig=\n' + ifConfigOut);
            }
            
            // Step 2: Get Route information
            cp.exec('/sbin/route', function (err, routeOut, stderr) {
                if (err || stderr) {
                    NetworkManagerLogger.error('NetworkManager.GetNetworkInterfaces: route err=' + err + ' stderr=' + stderr);
                } else {
                    NetworkManagerLogger.info('NetworkManager.GetNetworkInterfaces: raw /sbin/route=\n' + routeOut);
                }
                
                // Step 3: Get NameServers information
                cp.exec('cat /etc/resolv.conf', function (err, NameServersOut, stderr) {
                    if (err || stderr) {
                        NetworkManagerLogger.error('NetworkManager.GetNetworkInterfacesUsingIfconfig: resolv err=' + err + ' stderr=' + stderr);
                    } else {
                        NetworkManagerLogger.info('NetworkManager.GetNetworkInterfacesUsingIfconfig: raw /etc/resolv.conf=\n' + NameServersOut);
                    }
                    
                NetworkManagerLogger.info('NetworkManager.GetNetworkInterfaces: raw ifconfig=' + ifConfigOut);
                NetworkManagerLogger.info('NetworkManager.GetNetworkInterfaces: raw /sbin/route=' + routeOut);
                NetworkManagerLogger.info('NetworkManager.GetNetworkInterfaces: raw /etc/resolv.conf=' + NameServersOut);
                    var networkInterfaces = parse(ifConfigOut, routeOut, NameServersOut);
                NetworkManagerLogger.trace('NetworkManager.GetNetworkInterfaces: Parsed interfaces=' + JSON.stringify(networkInterfaces));
                    
                    for (var i = 0; i < networkInterfaces.length; i++) {
                        //NetworkManagerLogger.info('NetworkManager.GetNetworkInterfacesUsingIfconfig - Network Interface['+i+']='+JSON.stringify(networkInterfaces[i]))
                        
                        if (networkInterfaces[i].Name.indexOf('eth') < 0 || !networkInterfaces[i].Name) {
                            networkInterfaces.splice(i, 1)
                            i--;// Need to do this because we just removed i - I should fix this to make it push to a new array list, cuz this is yucky!
                        }
                    }
                    NetworkManagerLogger.info('NetworkManager.GetNetworkInterfaces: Parsed interfaces[Total:' + networkInterfaces.length + ']=' + JSON.stringify(networkInterfaces));

                    // Step 4: Find out the method (DHCP or static) for each network interface, add Method element
                    GetNetworkInterfaceMethod(networkInterfaces, function (err, networkInterfaces) {
                        
                        // Return the Network Interfaces object
                        callback(err, networkInterfaces);

                    });// Step 4: Get Interface Method
                });// Step 3: Get NameServers
            });// Step 2: Get Route
        });// Step 1: Get IfConfig    
    } catch (error) {
        // Error getting network interfaces information
        NetworkManagerLogger.error("NetworkManager.GetNetworkInterfaces: Error retreiving network information. Error=" + error);
        callback(error, null);
    }
}

// Return the method of the network interfaces
//#region Methods used to get network interfaces information
GetNetworkInterfaceMethod = function (networkInterfaces, callback) {
    
    //--- Read /etc/network/interfaces file
    cp.exec('cat /etc/network/interfaces', function (err, stdout, stderr) {
        if (err) {
            NetworkManagerLogger.error('NetworkManager.GetNetworkInterfaceMethod: /etc/network/interfaces err=' + err);
            callback(err, networkInterfaces);
        }
        
        if (stderr) {
            NetworkManagerLogger.error('NetworkManager.GetNetworkInterfaceMethod: /etc/network/interfaces stderr=' + stderr);
            callback(stderr, networkInterfaces);
        }
        
        // Break apart the /etc/network/interfaces file in preparation for analysis        
        var lines = stdout.split('\n');// Break in to separate lines
        //NetworkManagerLogger.trace('NetworkManager.GetNetworkInterfaceMethod: /etc/network/interfaces - # lines=' + lines.length);
        
        // Modify the method type within the Network Interfaces object
        for (i = 0; i < networkInterfaces.length; i++) {
            //NetworkManagerLogger.trace('NetworkManager.GetNetworkInterfaceMethod: /etc/network/interfaces - parsing for ' + networkInterfaces[i].Name);
            
            var method = "UNKNOWN";// Assume method is unknown, until we find out otherwise
            for (l = 0; l < lines.length; l++) {
                // Check to see if have line that begins with 'iface'
                //NetworkManagerLogger.trace('NetworkManager.GetNetworkInterfaceMethod: /etc/network/interfaces - line[' + l + ']=' + lines[l]);
                if (_.startsWith(lines[l], 'iface')) {
                    // This is a relevant line, let's break it apart to see if it matters to us
                    //NetworkManagerLogger.trace('NetworkManager.parseMethod: etcNetworkInterfaces - found line with iface=' + lines[l]);
                    methodInfo = lines[l].split(' ');
                    NetworkManagerLogger.info('NetworkManager.GetNetworkInterfaceMethod: /etc/Network/Interfaces interface=' + methodInfo[1] + ' method=' + methodInfo[3]);
                    if (methodInfo[1] === networkInterfaces[i].Name) {
                        method = _.upperCase(methodInfo[3]);
                    }
                }
            }
            networkInterfaces[i].Method = method;
        }
        
        // Return the adjusted Network Interfaces object
        callback(err, networkInterfaces);
    });// Get network interface method
}

function parse(ifConfigOut, routeOut, NameServersOut) {
    //NetworkManagerLogger.trace('NetworkManager.parse: ifConfigOut'); NetworkManagerLogger.trace(ifConfigOut);
    //NetworkManagerLogger.trace('NetworkManager.parse: routeOut'); NetworkManagerLogger.trace(routeOut);
    //NetworkManagerLogger.trace('NetworkManager.parse: NameServersOut'); NetworkManagerLogger.trace(NameServersOut);
    
    return ifConfigOut.split('\n\n').map(function (inface) {
        //NetworkManagerLogger.trace('NetworkManager.parse: inface'); NetworkManagerLogger.trace(inface);
        var lines = inface.split('\n');
        
        /**
         * Format 1
        * link xx:xx HWaddr xx-xx-xx
        * link xx:xx HWaddr xx:xx:xx
        *
        * Format 1
        * inet xx:xxx.xxx.xxx.xxx mask|mask|...:xxx.xxx.xxx.xxx
        */

        return {
            Name: getInterfaceName(_.first(lines)),
            IPAddress: getInterfaceIpAddr(lines[1]),
            Netmask: getInterfaceNetmaskAddr(lines[1]),
            NetworkStatus: getNetworkStatus(lines[3]),
            Broadcast: getBroadcastAddr(lines[1]),
            MACID: getInterfaceMacAddr(_.first(lines)),
            Gateway: getGateway(routeOut),
            NameServers: getNameServers(NameServersOut)
        };
    });
}

function getInterfaceName(firstLine) {
    return _.first(firstLine.split(' '));
}

function getInterfaceMacAddr(firstLine) {
    /**
 * extract mac adress
 *
 * ifconfig output:
 *   - link xx:xx HWaddr xx-xx-xx
 *   - link xx:xx HWaddr xx:xx:xx
 *
 * @param  {string} firstLine
 * @return {string}           Mac address, format: 'xx:xx:xx:xx:xx:xx'
 */
    if (!_.includes(firstLine, MAC)) {
        return null;
    }
    
    var macAddr = _.last(firstLine.split(MAC)).trim().replace(/-/g, ':');
    
    if (macAddr.split(':').length !== 6) {
        return null;
    }
    
    return macAddr;
}

function getInterfaceIpAddr(line) {
    /**
 * extract ip addr
 *
 * ifconfig output:
 *   - inet xx:xxx.xxx.xxx.xxx mask|mask|...:xxx.xxx.xxx.xxx
 *
 * @param  {string} line
 * @return {string,null} xxx.xxx.xxx.xxx
 */
    
    if (!_.includes(line, INET)) {
        return null;
    }
    return _.first(line.split(':')[1].split(' '));
}

function getInterfaceNetmaskAddr(line) {
    /**
 * extract netmask addr
 *
 * ifconfig output:
 *   - inet xx:xxx.xxx.xxx.xxx mask|mask|...:xxx.xxx.xxx.xxx
 *
 * @param  {string} line
 * @return {string,null} xxx.xxx.xxx.xxx
 */
    if (!_.includes(line, INET)) {
        return null;
    }
    return _.last(line.split(':'));
}

function getNetworkStatus(line) {
    /**
 * extract broadcast addr
 * @param  {string} line
 * @return {string,null}      xxx.xxx.xxx.xxx
 */

    // UP LOOPBACK RUNNING MTU:xxx Metric:yyy
    if (_.includes(line, UP)) {
        if (_.includes(line, RUNNING)) {
            return RUNNING;
        }
        else {
            return UNPLUGGED;
        }
    }
    else {
        return DOWN;
    }
}

function getBroadcastAddr(line) {
    /**
 * extract broadcast addr
 * @param  {string} line
 * @return {string,null}      xxx.xxx.xxx.xxx
 */
        if (!_.includes(line, BCAST)) {
        return null;
    }
    
    // inet adr:1.1.1.77  Bcast:1.1.1.255  Mask:1.1.1.0
    // @todo oh boy. this is ugly.
    return _.chain(line)
    .split(BCAST)
    .slice(1)
    .first()
    .value()
    .substring(1)
    .split(' ')[0];
}

function getGateway(routeOut) {
    /**
 * extract gateway ip
 * @param  {string} stdout
 * @return {string,null} default gateway ip or null
 */    
    var temp;
    try {
        temp = _.split(routeOut, '\n');// create 
        temp = _.filter(temp, function (line) {
            return _.includes(line, 'default');// Select only lines with 'default'

        });
        temp = _.head(temp);
        temp = _.split(temp, ' ');
        temp = _.without(temp, '');
        gateway = _.nth(temp, 1)
    }
    catch (error) {
        gateway = '';
    }
    finally {
        //NetworkManagerLogger.trace('NetworkManager.PutNetworkInterfaces: Gateway=' + gateway);
        return gateway;
    }
}

function getNameServers(NameServersOut) {
    /**
 * extract gateway ip
 * @param  {string} stdout
 * @return {string,null} default gateway ip or null
 */
    var temp;
    var NameServers = [];
    try {
        temp = _.split(NameServersOut, '\n');// Split each route line into array
        temp = _.filter(temp, function (line) {
            return _.includes(line, 'nameserver');// Select only lines with 'nameserver'
        });
        
        var tempNameServer;
        for (var i = 0; i < temp.length; i++) {
            tempNameServer = _.split(temp[i], ' ');// Split line into array
            tempNameServer = _.without(tempNameServer, '');// Get rid of unneed array elements
            tempNameServer = _.nth(tempNameServer, 1);// Grab the second element in the array which is the IP address
            NameServers.push(tempNameServer);// push the name server to the return value
        }
    }
    catch (error) {
        NameServers = [];
    }
    finally {
        //NetworkManagerLogger.trace('NetworkManager.PutNetworkInterfaces: NameServers=' + JSON.stringify(NameServers));
        return NameServers;
    }
}
//#endregion Methods used to get network interfaces information

//#endregion Get Network Interfaces

//#region Put Network Interfaces
exports.PutNetworkInterfaces = function (interfaces, callback) {
    //NetworkManagerLogger.trace('NetworkManager.PutNetworkInterfaces: interfaces=' + JSON.stringify(interfaces));
    var err = false;
    const exec = cp.exec;
    
    // Copy the interfaces file to a backup
    //#region Copy /etc/network/interfaces file to backup
    exec('cp /etc/network/interfaces /etc/network/interfaces.bak', function (error, stdout, stderr) {
        if (error) {
            // Error 'cp /etc/network/interfaces /etc/network/interfaces.bak'
            NetworkManagerLogger.error('NetworkManager.PutNetworkInterfaces: cp /etc/network/interfaces /etc/network/interfaces.bak. Error=' + error);
            callback(error, interfaces);
        }
        NetworkManagerLogger.info('NetworkManager.PutNetworkInterfaces: cp /etc/network/interfaces /etc/network/interfaces.bak. Success=' + stdout);
        //#endregion Copy /etc/network/interfaces file to backup
        
        //#region Define interfaces file content
        // We're going to now write the contents of the /etc/network/interfaces file
        var interfacesContent = '';// String that will hold output to file
        
        //  Preamble
        interfacesContent += '# interfaces(5) file used by ifup(8) and ifdown(8)\n\n';
        
        // lo
        //      We'll define the loopback interface here
        interfacesContent = '# loopback network interface\n';
        interfacesContent += 'auto lo\n';
        interfacesContent += 'iface lo inet loopback\n\n';
        
        // eth0
        //      We're going to configure eth0 as either DHCP or static, depending
        //      on what was passed within the interfaces structure.    
        if (interfaces.Method === 'DHCP') {
            interfacesContent += '# primary network interface\n';
            interfacesContent += 'auto ' + interfaces.Name + '\n';
            interfacesContent += 'iface ' + interfaces.Name + ' inet dhcp\n';
            interfacesContent += 'hwaddress ether ' + interfaces.MACID + '\n\n';
        } else if (interfaces.Method === 'STATIC') {
            interfacesContent += '# primary network interface\n';
            interfacesContent += 'auto ' + interfaces.Name + '\n';
            interfacesContent += 'iface ' + interfaces.Name + ' inet static\n';
            interfacesContent += 'address ' + interfaces.IPAddress + '\n';
            interfacesContent += 'netmask ' + interfaces.Netmask + '\n';
            if (interfaces.Gateway !== '') interfacesContent += 'gateway ' + interfaces.Gateway + '\n';
            // Don't write out empty name servers information
            var nameserverlength = 0;
            for (i = 0; i < interfaces.NameServers.length; i++) {
                if (interfaces.NameServers[i].length > 0) nameserverlength++;
            }
            if (nameserverlength) {
                interfacesContent += 'dns-nameservers ';
                for (i = 0; i < nameserverlength; i++) {
                    if (interfaces.NameServers[i].length > 0) interfacesContent += interfaces.NameServers[i] + ' ';
                }
                interfacesContent += '\n';
            }
            interfacesContent += 'hwaddress ether ' + interfaces.MACID + '\n\n';
        } else {
            err = true;// Huh?!? 
        }
        
        // eth1
        //      We'll need to fix this up once we actually have eth1 working but for now, we'll make it simple
        interfacesContent += '# secondary network interface\n';
        interfacesContent += 'auto eth1\n';
        interfacesContent += 'iface eth1 inet manual\n';
        // We'll add 1 to the eth0's MACID to get this MACID
        var macidString = interfaces.MACID.toString().split(':');
        macidString[5] = ("00" + (parseInt(macidString[5], 16) + 1).toString(16)).substr(-2);// Ensure we have preceding 0 if needed
        
        interfacesContent += 'hwaddress ether ' + macidString.join(':') + '\n';
        
        // Write the content to the /etc/network/interfaces file 
        NetworkManagerLogger.debug('NetworkManager.PutNetworkInterfaces: Data to write to /etc/network/interfaces file: ' + interfacesContent.toString());
        
        //#endregion Define interfaces file content
        
        //#region Save content to interfaces file
        fs.writeFile("/etc/network/interfaces", interfacesContent, function (error) {
            if (error) {
                NetworkManagerLogger.error('NetworkManager.PutNetworkInterfaces: fs.writeFile /etc/network/interfaces. Error=' + error);
                callback(error, interfaces);
            }
            NetworkManagerLogger.info('NetworkManager.PutNetworkInterfaces: fs.writeFile /etc/network/interfaces. Success=' + interfacesContent);
            
            // Ensure that the file contents are actually written to disk
            //  Without this command, the file contents may still be in the cache
            //  and when rebooted, might be lost
            exec("sync", function (error, stdout, stderr) {
                NetworkManagerLogger.info('sync stdout: ' + stdout);
                if (error) {
                    NetworkManagerLogger.error('NetworkManager.PutNetworkInterfaces: exec sync. Error: ' + error);
                }
                
                // Call back to finish up our configuration changes
                //  return back to calling API function
                callback(null, interfaces);
            });
        });
        //#endregion Save content to interfaces file
    });
}
//#endregion Put Network Interfaces
//#endregion Network Interfaces

//---------------------------------------------
// SNMP Server
//#region SNMP Server

var SnmpHostArray = [];
// Initialize SNMP Host array
exports.InitializeSnmpHostSettings = function (callback) {
    // Try getting the DigitalInputs table
    NetworkManagerDb.get("SELECT name FROM sqlite_master WHERE type='table' AND name='SnmpHosts'", function (err, rows) {
        if (err !== null) {
            NetworkManagerLogger.error('SNMP Manager - Initialize Error (Select)=' + JSON.stringify(err));
        }
        else if (typeof rows === "undefined") {
            // We read nothing, so let's create the table ...
            NetworkManagerLogger.info("SNMP Manager - Initialize by creating database table ...");
            NetworkManagerDb.exec("CREATE TABLE if not exists SnmpHosts (HostName TEXT, Enabled TEXT, IpAddress TEXT, Port TEXT, Community TEXT )", function (err) {
                if (err !== null) {
                    NetworkManagerLogger.error('SNMP Manager - Initialize Error (Create)=' + JSON.stringify(err));
                }
                else {
                    if (typeof callback == "function") {
                        NetworkManagerLogger.info('SNMP Manager - Initialized new database table');
                        callback(SnmpHostArray);
                    }
                }
            });
        }
        else {
            // Read from the database and populate the digitalInputsArray
            NetworkManagerLogger.info("SNMP Manager - Read SNMP Hosts from database ...");
            numberOfSnmpHosts = 0;
            SnmpHostArray.length = 0;
            NetworkManagerDb.each("SELECT rowid AS Id, HostName, Enabled, IpAddress, Port, Community FROM SnmpHosts", function (err, row) {
                if (err !== null) {
                    NetworkManagerLogger.error('SNMP Manager - Initialize Error (Select)=' + JSON.stringify(err));
                } else {
                    // Copy each row into a SNMPHostArray
                    NetworkManagerLogger.debug('SNMP Manager - Processing Host[' + (numberOfSnmpHosts + 1) + ']=' + row.HostName);
                    SnmpHostArray.push({
                        Id : row.Id,
                        HostName : row.HostName,
                        Enabled : row.Enabled,
                        IpAddress : row.IpAddress,
                        Port : row.Port,
                        Community : row.Community
                    });
                    NetworkManagerLogger.info("SNMP Manager - Read SNMP Hosts from Database[" + numberOfSnmpHosts + "]=" + JSON.stringify(SnmpHostArray[numberOfSnmpHosts]));
                    numberOfSnmpHosts++;
                }
            }, function () {
                // Retrurn the complete list of SNMP Hosts
                NetworkManagerLogger.info('SNMP Manager - Initialized from existing database table');
                if (typeof callback == "function") {
                    callback(SnmpHostArray);
                }
            });
        }
    });
}

// Create SNMP Host sessions
exports.CreateSnmpHostSessions = function (snmpHostArray, callback) {
    NetworkManagerLogger.info("SNMP Manager - Creating SNMP Hosts sessions ...");
    
    // Default SNMP session options
    var SnmpSessionOptions = {
        port: 161,
        retries: 1,
        timeout: 5000,
        transport: "udp4",
        trapPort: 162,
        version: NetworkManagerNetSnmp.Version2c
    };
    
    var snmpHostSessions = [];
    for (i = 0; i < snmpHostArray.length; i++) {
        SnmpHostAddress = snmpHostArray[i].IpAddress;
        SnmpCommunity = snmpHostArray[i].Community;
        SnmpOptions = JSON.parse(JSON.stringify(SnmpSessionOptions));
        SnmpOptions.trapPort = snmpHostArray[i].Port;
        snmpSession = NetworkManagerNetSnmp.createSession(SnmpHostAddress, SnmpCommunity, SnmpOptions);
        snmpHostSessions.push({
            Session : snmpSession,
            Enabled : snmpHostArray[i].Enabled
        });
        NetworkManagerLogger.info("SNMP Manager.CreateSnmpHostSessions - Created Snmp Host Session - Host[" + i + "]='" + snmpHostArray[i].HostName + "' @ " + SnmpHostAddress + ":" + SnmpOptions.trapPort + " (" + SnmpCommunity + ")");
    }
    
    if (typeof callback == "function") {
        callback(snmpHostSessions);
    }
}
//#endregion
